local AddonName, GGL = ...

----------------
-- Window for Challenge
----------------

local AceGUI = LibStub("AceGUI-3.0")

-- Lua APIs
local pairs, assert, type = pairs, assert, type
local tblsort,tblremove = table.sort, table.remove

-- WoW APIs
local PlaySound = PlaySound
local CreateFrame, UIParent = CreateFrame, UIParent
local C_TimerAfter = C_Timer.After

-- GGL
local L = GGL.Locale

-- config
local Type = "GoldGoblinLootGUIContainer-TabelView"
local LineType = "GoldGoblinLootGUIWidget-TabelViewLine"
local LINE_HEIGHT = 15
local Version = 1

--------------------------
-- table entry (widget) --
--------------------------
do
    local GAP_SIZE = 3
    local CellRegister = {
        ["text"] = {},
    }

    local CellCreation = {
        ["text"] = function()
            local frame = CreateFrame("Frame")
            frame:SetSize(10,10)
            frame.type = "text"

            local text = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
            text:SetAllPoints(frame)
            text:SetJustifyH("LEFT")
            text:SetJustifyV("MIDDLE")
            text:SetWordWrap(false)
            frame.text = text

            return frame
        end,
    }

    local function GetCell(cellType)
        local cell = next(CellRegister[cellType])
        if not cell then
            cell = CellCreation[cellType]()
        end
        CellRegister[cellType][cell] = nil
        cell:ClearAllPoints()
        cell:Show()
        return cell
    end

    local function FreeCell(cell)
        CellRegister[cell.type][cell] = true
        cell:ClearAllPoints()
        cell:Hide()
    end

    local methods = {
        ["OnAcquire"] = function(self)
            self:SetSortValue()
            self:SetAsHeader()
            self:SetContent()
            self:SetHeight(LINE_HEIGHT)
            self:SetFreeScaleCell()
            self:ShowBackground(false)
        end,

        ["OnRelease"] = function(self)

            -- release all cells
            self:ClearCells()
        end,

        ["OnWidthSet"] = function(self, width)

        end,

        ["OnHeightSet"] = function(self, height)
            if self._height and self._height == height then return end
            self._height = height
            for k, v in pairs(self.cells) do
                v:SetHeight(self._height)
            end
        end,

        ["ShowBackground"] = function(self, state)
            if state then
                self.frameBG:Show()
            else
                self.frameBG:Hide()
            end
        end,

        --- Sets the line as header line
        -- @param       state       <bool>
        ["SetAsHeader"] = function(self, state)
            self.isHeader = state
        end,

        --- Sets the cell thats scaleable, thats only for the header
        -- @param       index       <number>
        ["SetFreeScaleCell"] = function(self, index)
            self.freeScaleCell = index or 1
        end,

        --- Sets the content for that line
        -- @param       list        <TabelViewLine>
        -- @param       list        <table> index {"text1", "text2"}
        ["SetContent"] = function(self, header, list)
            if not header or not header.isHeader or not list then
                self:ClearCells()
                return
            end

            for cellIndex = 1, #list do
                -- check if a cellEntry exists
                if list[cellIndex] then
                    local cell = GetCell(self.cellType and (self.cellType[cellIndex] or "text") or "text")
                    cell:SetParent(self.frame)
                    cell:SetPoint("TOP", self.frame, 0, 0)
                    cell:SetPoint("LEFT", header.cells[cellIndex], 0, 0)
                    cell:SetPoint("RIGHT", header.cells[cellIndex], 0, 0)
                    cell:SetHeight(self._height)
                    cell.text:SetText(list[cellIndex])
                    if cell.text:GetUnboundedStringWidth() > header.cells[cellIndex]:GetWidth() then
                        header.cells[cellIndex]:SetWidth(cell.text:GetUnboundedStringWidth())
                    end
                    if cellIndex == 1 then
                        self.sortValue = list[cellIndex]
                    end
                    self.cells[cellIndex] = cell
                end
            end

            self.frame:Show()
        end,

        --- Adds a single cell
        -- @param       index       <number>
        -- @param       noReanchor  <bool> Call "ReanchorCells" when finished with adding!
        ["AddCell"] = function(self, index, noReanchor)
            if not self.isHeader then return end
            if self.cells[index] then return self.cells[index] end
            local cell = GetCell(self.cellType and (self.cellType[index] or "text") or "text")
            cell:SetParent(self.frame)
            cell:SetHeight(self._height)
            self.cells[index] = cell

            if not noReanchor then
                self:ReanchorCells()
            end

            return cell
        end,

        ["ReanchorCells"] = function(self)
            if not self.isHeader then return end
            -- first from left
            for i = 1, #self.cells do
                local cell = self.cells[i]

                if i == 1 then
                    cell:SetPoint("LEFT", self.frame, 0, 0)
                else
                    cell:SetPoint("LEFT", self.cells[i-1], "RIGHT", GAP_SIZE, 0)
                end

                -- if the currentLine is the freeScaleLine we need a new loop
                if i == self.freeScaleCell and i == #self.cells then
                    cell:SetPoint("RIGHT", self.frame, 0, 0)
                    return
                elseif i == self.freeScaleCell then
                    break
                end

                cell:SetWidth(cell.text:GetUnboundedStringWidth()+2)
            end

            -- second loop for right side
            for i = #self.cells, 1, -1 do
                local cell = self.cells[i]

                if i == #self.cells then
                    cell:SetPoint("RIGHT", self.frame, 0, 0)
                else
                    cell:SetPoint("RIGHT", self.cells[i+1], "LEFT", -(GAP_SIZE), 0)
                end

                -- if the currentLine is the freeScaleLine we need a new loop
                if i == self.freeScaleCell then
                    return
                else
                    cell:SetWidth(cell.text:GetUnboundedStringWidth()+2)
                end
            end
        end,

        --- Clears all cells from the line
        ["ClearCells"] = function(self)
            self.lineList = nil
            self.headerLine = nil
            for k, v in pairs(self.cells) do
                FreeCell(v)
                self.cells[k] = nil
            end
        end,

        ["SetSortValue"] = function(self, value)
            self.sortValue = value or 0
        end,

        ["GetSortValue"] = function(self)
            return self.sortValue or 0
        end,

        ["UpdateListRank"] = function(self, rank)
            self:ShowBackground(rank%2==0)
        end,
    }

    local function LineConstructor()
        local widget = {}
        widget.type  = LineType

        local frame = CreateFrame("Frame")
        widget.frame = frame

        local frameBG = frame:CreateTexture(nil, "BACKGROUND")
        frameBG:SetAllPoints(frame)
        frameBG:SetColorTexture(1, 1, 1, 0.2)
        frameBG:Hide()
        widget.frameBG = frameBG

        widget.cells = {}

        -- methodes
        for method, func in pairs(methods) do
            widget[method] = func
        end

        return AceGUI:RegisterAsWidget(widget)
    end

    AceGUI:RegisterWidgetType(LineType, LineConstructor, Version)
end

----------------
-- Main Frame --
----------------
--[[
    Events :

]]
local LineRegister = {}

local function CreateLine()
    local frame = CreateFrame("Frame")
    frame:SetSize(2,1)

    local line = frame:CreateTexture(nil, "BACKGROUND")
    line:SetAllPoints(frame)
    line:SetColorTexture(1, 1, 1, 1)

    return line
end

local function GetLine()
    local line = next(LineRegister)
    if not line then
        line = CreateLine()
    end
    LineRegister[line] = nil
    line:ClearAllPoints()
    line:Show()
    line:SetWidth(1)
    return line
end

local function FreeLine(line)
    LineRegister[line] = true
    line:ClearAllPoints()
    line:Hide()
end

local methods = {
    ["OnAcquire"] = function(self)
        self.scrollContainer:OnAcquire()
    end,

    ["OnRelease"] = function(self)
        self.scrollContainer:ReleaseChildren()
    end,

    ["OnWidthSet"] = function(self, width)
        --self.scrollContainer:SetWidth(width)
    end,

    ["OnHeightSet"] = function(self, height)
        --self.scrollContainer:OnHeightSet(height)
    end,

    ["LayoutFinished"] = function(self, width, height)
        --self.scrollContainer:LayoutFinished()
    end,

    ["OnRelease"] = function(self)
        self:SetContent()
    end,

    --- Sets the cell that gets free scaled
    -- @param       index       <number>
    ["SetFreeScaleCell"] = function(self, index)
        self.freeScaleCell = index
        self.headerLine:SetFreeScaleCell(index)
    end,

    --- removes all lines
    ["FreeLines"] = function(self)
        for i = 1, #self.lines do
            FreeLine(self.lines[i])
            self.lines[i] = nil
        end
    end,

    --- Sets the content for the
    -- @param       headerData  <table> {"title1", "title2", ...}
    -- @param       list        <table> {"cell1", "cell2", ...}
    ["SetContent"] = function(self, headerData, list)
        self.scrollContainer:ReleaseChildren()
        self.headerLine:ClearCells()
        self:FreeLines()
        if not headerData or not list then
            self.headerData, self.list = nil ,nil
            return
        end

        -- create header
        for i = 1, #headerData do
            local cell = self.headerLine:AddCell(i, true)
            cell.text:SetText(headerData[i])
        end
        self.headerLine:ReanchorCells()

        -- add Lines
        for i = 1, #self.headerLine.cells-1 do
            local line = GetLine()
            line:SetParent(self.frame)
            line:SetPoint("TOPLEFT", self.headerLine.cells[i], "TOPRIGHT", 1, 0)
            line:SetPoint("BOTTOM", self.scrollContainer.frame, 0, 0)
            self.lines[#self.lines+1] = line
        end

        -- add content
        self.scrollContainer:PauseLayout()
        for i = 1, #list do
            local cell = AceGUI:Create(LineType)
            cell:SetContent(self.headerLine, list[i])
            cell:SetFullWidth(true)
            self.scrollContainer:AddChild(cell)
        end
        self.scrollContainer:ResumeLayout()
        self.scrollContainer:SortChildrenList()

        self.headerData, self.list = headerData ,list
    end,
}

local function ExportButton_OnClick(self, event, button)
    if not self.obj or not self.obj.list then return end
    GGL.Export:FillExportWindow(self.obj.list, "csv")
end

local function Constructor()
    local frame = CreateFrame("FRAME")

    local container = {
        frame = frame,
        lines = {},
    }

    local exportButton = AceGUI:Create("Button")
    exportButton.frame:ClearAllPoints()
    exportButton.frame:SetParent(frame)
    exportButton.frame:SetPoint("TOPRIGHT", frame, 0, 0)
    exportButton:SetAutoWidth(true)
    exportButton:SetText(L["Export"])
    exportButton:SetCallback("OnClick", ExportButton_OnClick)
    exportButton.frame:Show()
    exportButton.obj = container
    container.exportButton = exportButton

    local headerLine = AceGUI:Create(LineType)
    headerLine.frame:ClearAllPoints()
    headerLine.frame:SetParent(frame)
    headerLine:SetAsHeader(true)
    headerLine.frame:SetPoint("TOP", exportButton.frame, "BOTTOM", 0, -2)
    container.headerLine = headerLine

    local headerBottomLine = GetLine()
    headerBottomLine:SetParent(frame)
    headerBottomLine:SetHeight(1)
    headerBottomLine:SetPoint("TOPLEFT", headerLine.frame, "BOTTOMLEFT", 0, -1)
    headerBottomLine:SetPoint("TOPRIGHT", headerLine.frame, "BOTTOMRIGHT", 0, -1)

    local scrollContainer = AceGUI:Create("GoldGoblinLootGUIContainer-SortedScrollFrame")
    scrollContainer.frame:ClearAllPoints()
    scrollContainer.frame:SetParent(frame)
    scrollContainer:SetFullWidth(true)
    scrollContainer:SetPoint("TOPRIGHT", exportButton.frame, "BOTTOMRIGHT", 0, -(headerLine.frame:GetHeight() + 5))
    scrollContainer:SetPoint("BOTTOMLEFT", frame, 0, 0)
    scrollContainer:SetLayout("List")
    container.scrollContainer = scrollContainer

    headerLine.frame:SetPoint("LEFT", scrollContainer.content, 0, 0)
    headerLine.frame:SetPoint("RIGHT", scrollContainer.content, 0, 0)

    local content = CreateFrame("Frame", nil, frame)
    content:SetPoint("TOPLEFT", frame, 0, 0)
    content:SetPoint("BOTTOMRIGHT", exportButton.frame, "BOTTOMLEFT", 0, 0)
    container.content = content

    -- methods
    for k, v in pairs(methods) do
        container[k] = v
    end

    AceGUI:RegisterAsContainer(container)
    return container
end

AceGUI:RegisterWidgetType(Type,Constructor,Version)