local AddonName, GGL = ...

local L = GGL.Locale

-- lua
local type = type
local strsub, strtrim, str_match, format, strfind = string.sub, strtrim, string.match, string.format, string.find
local tblremove = table.remove

-- wow
local C_TimerAfter = C_Timer.After

-- other

-- ########################################
-- Base
-- ########################################

--- Formats a players display name
-- @param		name		<string>
-- @return	formatedName
function GGL.FormatDisplayName(name)
    name = strtrim(name)
	return strsub(name or "", 1, GGL.MAX_NAME_LENGTH)
end

-- ########################################
-- Fonts
-- ########################################
local FontRegister = {}

function GGL.GetFont(size, flags)
    if FontRegister[size..(flags or "")] then
        return FontRegister[size..(flags or "")]
    else
        local font = CreateFont("GGL-FONT-"..size..(flags or ""))
        font:SetFont(GGL.FONT_PATH, size, flags)
        FontRegister[size..(flags or "")] = font
        return font
    end
end

-- ########################################
-- LootMob (FastLoot)
-- ########################################
local LootMob_LoopRunning = false
local function LootMob_LootLoop()
    local numLoot = GetNumLootItems()
	if numLoot > 0 then
        for slot = 1, numLoot do
            LootSlot(slot)
            numLoot = numLoot - 1
        end
    end

    if numLoot > 1 then
        if not LootMob_LoopRunning then
            LootMob_LoopRunning = true
            C_TimerAfter(0.01, LootMob_LootLoop)
        end
    else
        LootMob_LoopRunning = false
        CloseLoot()
    end
end

--- Starts looting the open loot window
function GGL:LootMob()
    if GGL.db.Settings.FastAutoLoot and not LootMob_LoopRunning then
        LootMob_LootLoop()
    end
end

-- ########################################
-- DestroyGrayBagItems
-- ########################################

--- Destroys all grey items in the bags
-- @param       silent      <bool> print out chat msg
function GGL:DestroyGrayBagItems(silent)
    local counter = 0

    for bag = 0, NUM_BAG_SLOTS do
        for slot = 1, GetContainerNumSlots(bag) do
            local icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID = GetContainerItemInfo(bag, slot)
			if itemID and not locked and quality == 0 then
                PickupContainerItem(bag, slot)
                DeleteCursorItem()
                counter = counter + 1
			end
        end
    end

    if not silent and counter > 0 then
        GGL:FormatPrint(L["%d gray items destroyed."], counter)
    end

    return counter
end

-- ########################################
-- Timer
-- #########################################
local TIMER_PROTO = {
    --- Starts a timer if its not registered
    -- @param       interval        <number> in seconds
    -- @param       callback        <function>
    ["After"] = function(self, interval, callback)
        if interval and callback and not self.register[callback] then
            self.register[callback] = true
            C_TimerAfter(interval, callback)
        end
    end,

    --- This starts a new timer that runs as long the callback returns true
    -- @param       interval        <number> in seconds
    -- @param       callback        <function> resumes timer as long it returns true
    -- @callback    elapsed         <number> calls * interval
    ["ManagedAfter"] = function(self, interval, callback)
        if interval and callback and not self.register[callback] then
            local calls = 0
            local function managedAfter()
                calls = calls + 1
                if callback(calls*interval) then
                    C_TimerAfter(interval, managedAfter)
                else
                    self:Unregister(managedAfter)
                end
            end
            C_TimerAfter(interval, managedAfter)
            self.register[managedAfter] = callback
        end
    end,

    --- Unregister a callback from the list so it can set again
    -- @param       callback        <function>
    ["Unregister"] = function(self, callback)
        if callback and self.register[callback] then
            self.register[callback] = nil
            return true
        end
        return false
    end,
}
local TIMER_PROTO_MT = {__index = TIMER_PROTO}

--- Returns a table with an enhanced C_Timer
--- See TIMER_PROTO for more info
function GGL:GetTimer()
    local timer = {
        register = {}
    }

    setmetatable(timer, TIMER_PROTO)

    return timer
end

-- ########################################
-- CachTable
-- #########################################
local WEAK_TABLE_MT = {__mode = "kv"}

--- Gets a cache table
function GGL:GetCacheTable()
    return setmetatable({}, WEAK_TABLE_MT)
end

-- ########################################
-- ItemLink
-- #########################################
local ItemLinkCache = GGL:GetCacheTable()
local ITEM_LOG_FORMAT = "(%d+),*(.*)"
local ITEM_LINK_FORMT = "item:%d::::::::::::%s"
--/run print(GetItemInfo("item:3792::::::::::::1:6656:2:28:217:9:30:::"))
--- Formats item data get from Loot module
-- @return itemID, itemString
function GGL:GetItemDataFromItemLog(itemData)
    if not itemData then return end
    if ItemLinkCache[itemData] then
        return ItemLinkCache[itemData][1], ItemLinkCache[itemData][2]
    elseif type(itemData) == "string" then
        local itemID, itemBonus = str_match(itemData, ITEM_LOG_FORMAT)
        itemID = tonumber(itemID)
        if itemBonus and itemBonus ~= "" then
            itemBonus = format(ITEM_LINK_FORMT, itemID, itemBonus or "")
        else
            itemBonus = nil
        end
        ItemLinkCache[itemData] = {itemID, itemBonus}
        return itemID, itemBonus
    else
        return itemData
    end
end

-- ########################################
-- Instance Reset
-- Events
--		"OnInstancesResetNewLockout"	(instanceLockEntry, instanceLocks)
--		"OnInstancesResetUpdate"	(instanceLocks, nextReleaseStartTime, nextReleaseEndTimes)
--      "OnInstancesResetEnterInstance"     (enterCacheEntry, enterCache)
-- ########################################
local RESET_PATTERN =_G.INSTANCE_RESET_SUCCESS:gsub("%%s", "(.+)")
local RESET_UPDATE_TICK = 1
local MAX_RESETS = 10
local LOCK_TIME = 3600 -- 1h
local CACHE_TIME = 3600*2 -- 2h
local activeResetTimer, resetTimerIsRunning = false, false
local nextReleaseStartTime, nextReleaseEndTime

local function InstanceResetUpdateCache()
    local enterCache = GGL.db.InstanceResets.enterCache
    local curTime = GGL.GetTime()

    for i = 1, #enterCache do
        if (curTime - enterCache[i].time) > CACHE_TIME then
            tblremove(enterCache, i)
            return InstanceResetUpdateCache()
        end
    end
end

local function InstanceReset_OnUpdate(self, e)
    if not activeResetTimer then
        resetTimerIsRunning = false
        return
    end

	if nextReleaseEndTime <= GGL.GetTime() then
		GGL:InstanceResets_Update()
    end
    C_TimerAfter(RESET_UPDATE_TICK, InstanceReset_OnUpdate)
end

local function AddInstanceReset(zone)
    local instanceLocks = GGL.db.InstanceResets.locks
    local enterCache = GGL.db.InstanceResets.enterCache

    local enterTab

    -- check for enterTime
    for i = 1, #enterCache do
        if strfind(zone, enterCache[i].mapName) then
            enterTab = tblremove(enterCache, i)
            break
        end
    end

	instanceLocks[#instanceLocks + 1] = {
        zone = zone,
        enterTime = GGL.GetTime(),
		startTime = GGL.GetTime(),
        endTime = GGL.GetTime() + LOCK_TIME,
    }
    if enterTab then
        instanceLocks[#instanceLocks].mapID = enterTab.mapID
        instanceLocks[#instanceLocks].enterTime = enterTab.time or GGL.GetTime()
    end

    GGL:FireCallbackEvent("OnInstancesResetNewLockout", nil, instanceLocks[#instanceLocks], instanceLocks)
	GGL:InstanceResets_Update()
end

local function InstanceResets_OnMapChange(self, mapID, mapName, mapTypeFlag)
    if mapTypeFlag ~= GGL.MAP_TYPE_FLAGS.DUNGEON and mapTypeFlag ~= GGL.MAP_TYPE_FLAGS.RAID then return end

    InstanceResetUpdateCache()

    local enterCache = GGL.db.InstanceResets.enterCache

    for i = 1, #enterCache do
        if enterCache[i].mapID == mapID then
            return
        end
    end

    enterCache[#enterCache + 1] = {
        time = GGL.GetTime(),
        mapID = mapID,
        mapName = mapName,
    }

    GGL:FireCallbackEvent("OnInstancesResetEnterInstance", nil, enterCache[#enterCache], enterCache)
end

function GGL:CHAT_MSG_SYSTEM(msg)
	local zone = strmatch(msg, RESET_PATTERN)
	if zone then
		AddInstanceReset(zone)
	end
end

GGL:RegisterEvent("CHAT_MSG_SYSTEM")
GGL:RegisterCallbackEvent("OnMapChange", InstanceResets_OnMapChange)

--- Updates all resets
function GGL:InstanceResets_Update()
	local instanceLocks = GGL.db.InstanceResets.locks
	if #instanceLocks <= 0 then
		activeResetTimer = false
		GGL:FireCallbackEvent("OnInstancesResetUpdate", nil, instanceLocks, nil, nil)
		return
	else
		for i = 1, #instanceLocks do
			if instanceLocks[i].endTime <= GGL.GetTime() then
				local remove = tblremove(instanceLocks, i)
                nextReleaseEndTime = nil
                GGL:FireCallbackEvent("OnInstancesResetLockRemoves", nil, remove)
				GGL:InstanceResets_Update()
				return
			end
			if not nextReleaseEndTime or nextReleaseEndTime > instanceLocks[i].endTime then
				nextReleaseStartTime = instanceLocks[i].startTime
				nextReleaseEndTime = instanceLocks[i].endTime
			end
		end
        activeResetTimer = true
        if not resetTimerIsRunning then
            resetTimerIsRunning = true
            C_TimerAfter(RESET_UPDATE_TICK, InstanceReset_OnUpdate)
        end
	end
	GGL:FireCallbackEvent("OnInstancesResetUpdate", nil, instanceLocks, nextReleaseStartTime, nextReleaseEndTime)
end

--- Gets a table with alle entered instances since x
-- @param       since               <number:time> (optional) get all instance enters since that time
-- @param       addCurrentZone      <bool> adds the current zone when since is set
function GGL:InstanceResets_GetInstanceEnterTimes(since, addCurrentZone)
    if since then
        local enterCache = GGL.db.InstanceResets.enterCache
        local curZone = GGL:GetCurrentMapID()
        local ret = {}

        for i = 1, #enterCache do
            if (enterCache[i].time >= since) or (curZone and enterCache[i].mapID == curZone) then
                ret[#ret+1] = enterCache[i]
            end
        end

        return ret
    else
        return GGL.db.InstanceResets.enterCache
    end
end

--- Returns the InstanceResetTable with startTime since x
-- @param       since       <number:time> (optional) get all instance enters since that time
function GGL:InstanceResets_GetInstanceResetTimes(since)
    if since then
        local instanceLocks = GGL.db.InstanceResets.locks
        local ret = {}

        for i = 1, #instanceLocks do
            if instanceLocks[i].startTime >= since then
                ret[#ret+1] = instanceLocks[i]
            end
        end

        return ret
    else
        return GGL.db.InstanceResets.locks
    end
end
